"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SyncEvent = exports.SyncError = exports.SyncComplete = exports.SyncDenied = exports.SyncActive = exports.SyncPaused = exports.SyncChange = exports.SyncDB = exports.SyncManager = void 0;
//@flow
const zdb_1 = __importDefault(require("./zdb"));
const authBridge_1 = require("./authBridge");
const events_1 = require("events");
const pdb_1 = require("./pdb");
const SyncChange = "_sync_change";
exports.SyncChange = SyncChange;
const SyncPaused = "_sync_paused";
exports.SyncPaused = SyncPaused;
const SyncActive = "_sync_active";
exports.SyncActive = SyncActive;
const SyncDenied = "_sync_denied";
exports.SyncDenied = SyncDenied;
const SyncComplete = "_sync_complete";
exports.SyncComplete = SyncComplete;
const SyncError = "_sync_error";
exports.SyncError = SyncError;
function SyncEvent(db, event) {
    return db + "|" + event.toString();
}
exports.SyncEvent = SyncEvent;
class SyncManager {
    constructor(config, authBridge, debug = false) {
        this.getRemoteSyncURL = (db) => {
            const url = this._config.syncUrl + "/" + db;
            return url;
        };
        this._emit = (dbName, event, payload, error = false) => {
            this._eventEmitter.emit(SyncEvent(dbName, event), payload, error);
        };
        this._syncDB = (dbName) => {
            if (this._dbg)
                console.debug("_syncDB", dbName);
            const localDb = new zdb_1.default(dbName, { adapter: "react-native-sqlite" });
            let syncType = SyncType.UPSTREAM;
            if (this._config.download.includes(dbName)) {
                syncType = SyncType.DOWNSTREAM;
            }
            else if (this._config.upload.includes(dbName)) {
                syncType = SyncType.UPSTREAM;
            }
            else if (this._config.twoway.includes(dbName)) {
                syncType = SyncType.TWO_WAY;
            }
            else {
                console.warn("No sync DB defined for name " +
                    dbName +
                    ". Please make sure you have downloaded the latest config from the cloud, and this database is defined there." +
                    " (zest-auth.js)");
            }
            const sync = new SyncDB(this.getRemoteSyncURL, this._emit, dbName, localDb, this._authBridge.getAuthToken, syncType, this._dbg);
            return sync;
        };
        this.syncDB = (dbName) => {
            var _a;
            if (this._dbg)
                console.debug(" ==> SyncDB", dbName);
            if (this._syncDBs.has(dbName)) {
                return (_a = this._syncDBs.get(dbName)) === null || _a === void 0 ? void 0 : _a.getLocalDB();
            }
            console.warn("No sync DB defined for name " +
                dbName +
                ". Please make sure you have downloaded the latest config from the cloud, and this database is defined there." +
                " (zest-auth.js)", this);
            return undefined;
        };
        this.resetSyncDB = (dbName) => {
            if (this._dbg)
                console.debug(" ==> resetSyncDB", dbName);
            if (this._syncDBs.has(dbName)) {
                const sdb = this._syncDBs.get(dbName);
                if (sdb && sdb !== null) {
                    sdb.resetLocalDatabase();
                    return;
                }
                if (this._dbg)
                    console.debug(" ==> resetSyncDB failed to fetch", dbName);
            }
            else {
                if (this._dbg)
                    console.debug(" ==> resetSyncDB cannot find ", dbName);
            }
        };
        this.restartSync = (dbName) => {
            if (this._dbg)
                console.debug(" ==> restartSync", dbName);
            if (this._syncDBs.has(dbName)) {
                const sdb = this._syncDBs.get(dbName);
                if (sdb && sdb !== null) {
                    sdb.restartSync();
                    return;
                }
                if (this._dbg)
                    console.debug(" ==> restartSync failed to fetch", dbName);
            }
            else {
                if (this._dbg)
                    console.debug(" ==> restartSync cannot find ", dbName);
            }
        };
        this.getLocalDB = (dbName) => {
            if (this._localDBs.has(dbName)) {
                return this._localDBs.get(dbName);
            }
            console.warn("No local DB defined for name " +
                dbName +
                ". Please make sure you have added the correct local configuration." +
                " (zest-auth.js)", this);
            return undefined;
        };
        this.onUserLogin = (user) => {
            if (this._dbg)
                console.debug("onUserLogin", JSON.stringify(user));
            this._loggedIn = true;
            this.startSync();
        };
        this.onUserLogout = () => {
            this._loggedIn = false;
            this.stopSync();
        };
        this.startSync = () => {
            const sdbs = this._syncDBs;
            const keys = sdbs.entries();
            if (this._dbg)
                console.debug("startSync", keys);
            sdbs.forEach((value, key, map) => {
                value.startSync();
            }, this);
        };
        this.stopSync = () => {
            const sdbs = this._syncDBs;
            const keys = sdbs.entries();
            if (this._dbg)
                console.debug("startSync", keys);
            sdbs.forEach((value, key, map) => {
                value.stopSync();
            }, this);
        };
        this.getSyncConfig = () => {
            return this._config;
        };
        this._config = config;
        this._eventEmitter = new events_1.EventEmitter();
        this._authBridge = authBridge;
        this._loggedIn = false;
        this._dbg = debug;
        this._authBridge
            .on(authBridge_1.AuthLoggedIn, this.onUserLogin)
            .on(authBridge_1.AuthLoggedOut, this.onUserLogout);
        const dbs = new Map();
        if (this._config.download) {
            for (let db of this._config.download) {
                const s = this._syncDB(db);
                dbs.set(db, s);
                if (this._loggedIn) {
                    s.startSync();
                }
            }
        }
        if (this._config.upload) {
            for (let db of this._config.upload) {
                const s = this._syncDB(db);
                dbs.set(db, s);
                if (this._loggedIn) {
                    s.startSync();
                }
            }
        }
        if (this._config.twoway) {
            for (let db of this._config.twoway) {
                const s = this._syncDB(db);
                dbs.set(db, s);
                if (this._loggedIn) {
                    s.startSync();
                }
            }
        }
        this._syncDBs = dbs;
        // Local dabatabases that are not synced with backend
        const ldbs = new Map();
        if (this._config.local) {
            for (let db of this._config.local) {
                const localDb = new zdb_1.default(db, { adapter: "react-native-sqlite" });
                ldbs.set(db, localDb);
            }
        }
        this._localDBs = ldbs;
        if (this._dbg)
            console.debug("=> SyncManager initialized", this._syncDBs);
    }
    // *********************************
    // Add on(SyncEvent(myDBName, SyncChange), func(payload, error))
    // Add on(SyncChange, func(dbName, payload, error))
    // Add on(dbName, func(event, payload, error))
    on(eventName, listener) {
        this._eventEmitter.on(eventName, listener);
        return this;
    }
    removeListener(eventName, listener) {
        this._eventEmitter.removeListener(eventName, listener);
    }
}
exports.SyncManager = SyncManager;
const SyncType = {
    UPSTREAM: "up",
    DOWNSTREAM: "down",
    TWO_WAY: "2way",
};
class SyncDB {
    constructor(remoteUrlFunction, parentEmitter, name, localDb, getAuthTokenFunc, syncType, debug = false) {
        this.getLocalDB = () => {
            return this._localDb;
        };
        this.resetLocalDatabase = () => {
            this.stopSync();
            this._localDb.destroy().then(() => {
                this._localDb = new zdb_1.default(this._name, { adapter: "react-native-sqlite" });
                this.startSync();
            });
        };
        this.startSync = () => {
            if (this._dbg)
                console.debug("Start " + this._name + " " + this._syncType + " sync ");
            const gat = this._getAuthTokenFunc;
            const remoteDbURL = this._remoteUrlFunction(this._name);
            var handler;
            const obj = this;
            const remoteDb = new pdb_1.PouchDB(remoteDbURL, {
                fetch: function (url, opts) {
                    if (obj._dbg)
                        console.debug("=>", url, gat);
                    try {
                        const token = gat();
                        opts.headers.set("Authorization", token);
                        if (obj._dbg)
                            console.log("=> set auth: ", token);
                    }
                    catch (err) {
                        console.log("failed to get auth token", err);
                    }
                    return pdb_1.PouchDB.fetch(url, opts);
                },
            });
            try {
                switch (this._syncType) {
                    case SyncType.DOWNSTREAM:
                        handler = this._localDb.replicateFrom(remoteDb, {
                            live: true,
                            retry: true,
                            checkpoint: "target",
                        });
                        break;
                    case SyncType.UPSTREAM:
                        handler = this._localDb.replicateTo(remoteDb, {
                            live: true,
                            retry: true,
                            checkpoint: "source",
                        });
                        break;
                    case SyncType.TWO_WAY:
                        handler = this._localDb.sync(remoteDb, {
                            live: true,
                            retry: true,
                            checkpoint: "source",
                        });
                        break;
                    default:
                        throw Error("Failed to determine sync type" + this._syncType);
                }
            }
            catch (err) {
                console.error("Failed to setup replication", err);
            }
            if (handler) {
                this._handlerSync = handler
                    .on("change", (info) => {
                    if (this._dbg)
                        console.debug("change", this._name);
                    this._parentEmitter(this._name, SyncChange, info);
                })
                    .on("paused", (err) => {
                    if (this._dbg)
                        console.debug("paused", this._name);
                    this._parentEmitter(this._name, SyncPaused, err);
                })
                    .on("active", () => {
                    if (this._dbg)
                        console.debug("active", this._name);
                    this._parentEmitter(this._name, SyncActive, {});
                })
                    .on("denied", (err) => {
                    if (this._dbg)
                        console.debug("denied", this._name);
                    this._parentEmitter(this._name, SyncDenied, err);
                })
                    .on("complete", (info) => {
                    if (this._dbg)
                        console.debug("complete", this._name);
                    this._parentEmitter(this._name, SyncComplete, info);
                })
                    .on("error", (err) => {
                    if (this._dbg)
                        console.debug("error", this._name);
                    this._parentEmitter(this._name, SyncError, err);
                });
            }
        };
        this.restartSync = () => {
            this.stopSync();
            this.startSync();
        };
        this.stopSync = () => {
            if (this._handlerSync) {
                if (this._dbg)
                    console.debug("Stop " + this._name + " sync");
                this._handlerSync.cancel();
            }
        };
        this._dbg = debug;
        if (this._dbg)
            console.debug("[SyncDB] constructor", getAuthTokenFunc());
        this._remoteUrlFunction = remoteUrlFunction;
        this._parentEmitter = parentEmitter;
        this._localDb = localDb;
        this._name = name;
        this._getAuthTokenFunc = getAuthTokenFunc;
        this._syncType = syncType;
    }
}
exports.SyncDB = SyncDB;
